"""This module provides the dependencies for sthali-db usage.

Classes:
    PaginateParameters: Represents the parameters for retrieving items.

Functions:
    filter_parameters: Not implemented. Raises NotImplementedError.
"""

import typing

import pydantic


async def filter_parameters() -> typing.NoReturn:
    """Not implemented."""
    raise NotImplementedError


class PaginateParameters(pydantic.BaseModel):
    """Represents the parameters for retrieving items.

    Attributes:
        skip (pydantic.NonNegativeInt): The number of items to skip. Defaults to 0.
        limit (pydantic.NonNegativeInt): The maximum number of items to return. Defaults to 100.
    """

    skip: typing.Annotated[
        pydantic.NonNegativeInt,
        pydantic.Field(default=0, description="The number of items to skip"),
    ]
    limit: typing.Annotated[
        pydantic.NonNegativeInt,
        pydantic.Field(default=100, description="The maximum number of items to return"),
    ]



import unittest

import sthali_db.dependencies

module = sthali_db.dependencies


class TestFilterParameters(unittest.IsolatedAsyncioTestCase):
    async def test_not_implemented(self) -> None:
        with self.assertRaises(NotImplementedError):
            await module.filter_parameters()


class TestPaginateParameters(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = module.PaginateParameters()  # type: ignore

        self.assertEqual(result.skip, 0)
        self.assertEqual(result.limit, 100)

    async def test_return_custom(self) -> None:
        result = module.PaginateParameters(skip=10, limit=10)

        self.assertEqual(result.skip, 10)
        self.assertEqual(result.limit, 10)



import unittest

import sthali_db.types

module = sthali_db.types


class TestTypes(unittest.IsolatedAsyncioTestCase):
    async def test_get(self) -> None:
        types = module.Types()

        result = types.get("any")
        self.assertEqual(result, module.typing.Any)

    async def test_get_raise_exception(self) -> None:
        types = module.Types()

        with self.assertRaises(AttributeError):
            types.get("custom")

    async def test_set(self) -> None:
        types = module.Types()
        types.set("custom", module.typing.Any)

        result = types.get("custom")
        self.assertEqual(result, module.typing.Any)

    async def test_set_raise_exception(self) -> None:
        types = module.Types()
        types.set("custom", module.typing.Any)

        with self.assertRaises(TypeError):
            types.set("custom", module.typing.Any)

    async def test_pop(self) -> None:
        types = module.Types()
        types.pop("any")



"""This module provides a high-level interface for interacting with different database clients.

Classes:
    DB: Represents a database client adapter.

Dataclasses:
    DBSpecification: Represents the specification for a database connection.
"""

import pathlib
import typing

import pydantic
import sthali_core

if typing.TYPE_CHECKING:
    from .clients import Base


parent_path = pathlib.Path(__file__).parent
enum_clients_config = sthali_core.enum_clients.Clients(parent_path)
ClientEnum = enum_clients_config.enum


@pydantic.dataclasses.dataclass
class DBSpecification:
    """Represents the specification for a database connection.

    Attributes:
        path (str): Path to the database.
            This field specifies the path to the database file or server.
        client (str): One of available database clients.
            This field specifies the database client to be used for the connection.
            The available options are "Default", "Postgres", "Redis", "SQLite", and "TinyDB".
            Defaults to "Default".
    """

    path: typing.Annotated[str, pydantic.Field(description="Path to the database")]
    client: typing.Annotated[ClientEnum, pydantic.Field(description="One of available database clients")]


class DB:
    """Represents a database client adapter.

    Args:
        db_spec (DBSpecification): The specification for the database connection.
        table (str): The name of the table to interact with.
    """

    def __init__(self, db_spec: DBSpecification, table: str) -> None:
        """Initialize the DB instance.

        Args:
            db_spec (DBSpecification): The specification for the database connection.
            table (str): The name of the table to interact with.
        """
        client_name = str(db_spec.client.value)
        client_module = enum_clients_config.clients_map[client_name]
        client_class: type[Base] = getattr(client_module, f"{client_name.title()}Client")
        client = client_class(db_spec.path, table)

        self.insert_one = client.insert_one
        self.select_one = client.select_one
        self.update_one = client.update_one
        self.delete_one = client.delete_one
        self.select_many = client.select_many



"""This module provides the necessary components for interacting with the database."""

from .db import DB, DBSpecification
from .dependencies import PaginateParameters
from .models import FieldSpecification, Models
from .types import Types

__all__ = [
    "DB",
    "DBSpecification",
    "FieldSpecification",
    "Models",
    "PaginateParameters",
    "Types",
]



"""This module provides classes for creating dynamic models based on field specifications.

Classes:
    Models(name: str, fields: list[FieldSpecification]): Represents a collection of models.

Dataclasses:
    FieldSpecification: Represents a field with its metadata.
"""

import collections.abc
import typing
import uuid

import pydantic


@pydantic.dataclasses.dataclass
class FieldSpecification:
    """Represents a field with its metadata.

    Attributes:
        name (str): Name of the field.
        type (typing.Any): Type annotation of the field.
        default (Default | None): Default value/factory of the field. Defaults to None.
        description (str | None): Description of the field. Defaults to None.
        optional (bool | None): Indicates if the field accepts None. Defaults to None.
        title (str | None): Title of the field. Defaults to None.
    """

    @pydantic.dataclasses.dataclass
    class Default:
        """Represents a default value for an attribute.

        Attributes:
            factory (collections.abc.Callable[..., typing.Any] | None): The function used to create the default value
                forthe attribute. Defaults to None.
            value (typing.Any | None): The default value for the attribute. Defaults to None.
        """

        factory: typing.Annotated[
            collections.abc.Callable[..., typing.Any] | None,
            pydantic.Field(
                default=None, description="The function used to create the default value for the attribute",
            ),
        ]
        value: typing.Annotated[
            typing.Any | None,
            pydantic.Field(default=None, description="The default value for the attribute"),
        ]

    name: typing.Annotated[str, pydantic.Field(description="Name of the field")]
    type: typing.Annotated[typing.Any, pydantic.Field(description="Type annotation of the field")]
    default: typing.Annotated[
        Default | None,
        pydantic.Field(default=None, description="Default value/factory of the field"),
    ]
    description: typing.Annotated[str | None, pydantic.Field(default=None, description="Description of the field")]
    optional: typing.Annotated[
        bool | None,
        pydantic.Field(default=None, description="Indicates if the field accepts None"),
    ]
    title: typing.Annotated[str | None, pydantic.Field(default=None, description="Title of the field")]

    @property
    def _metadata(self) -> dict[str, typing.Any]:
        result: dict[str, typing.Any] = {
            "description": self.description or f"Field {self.name}",
            "title": self.title or self.name,
        }
        if self.default:
            if self.default.factory:
                result["default_factory"] = self.default.factory
            else:
                result["default"] = self.default.value
        return result

    @property
    def type_annotated(self) -> typing.Annotated[typing.Any, pydantic.Field]:
        """Returns the type annotation of the field.

        Returns:
            typing.Annotated[typing.Any, Field]: The type annotation of the field.
        """
        field_type = (self.type, typing.Union[self.type, None])[bool(self.optional)]
        return typing.Annotated[field_type, pydantic.Field(**self._metadata)]


class Models:
    """Represents a collection of models.

    This class is responsible for creating and managing models dynamically based on the provided fields.
    It provides methods to create different types of models such as create, response, and update models.

    Attributes:
        name (str): The name of the collection of models.
        create_model (type[Base]): The dynamically created model for creating new instances.
        response_model (type[BaseWithId]): The dynamically created model for response payloads.
        update_model (type[Base]): The dynamically created model for updating existing instances.
    """

    class Base(pydantic.BaseModel):
        """Represents a base class for models."""

    class BaseWithId(Base):
        """Represents a base class for models with a resource identifier."""

        id: typing.Annotated[uuid.UUID, pydantic.Field(description="Resource identifier")]

    def __init__(self, name: str, fields: list[FieldSpecification]) -> None:
        """Initialize the Models class.

        Args:
            name (str): The name of the collection of models.
            fields (list[FieldSpecification]): The list of fields specification for the models.
        """
        self.name = name
        self.create_model = self._factory(self.Base, f"Create{name.title()}", fields)
        self.response_model = self._factory(self.BaseWithId, f"Response{name.title()}", fields)
        self.update_model = self._factory(self.Base, f"Update{name.title()}", fields)

    @staticmethod
    def _factory(
        base: type[pydantic.main.ModelT],
        name: str,
        fields: list[FieldSpecification],
    ) -> type[pydantic.main.ModelT]:
        fields_constructor = {field.name: field.type_annotated for field in fields}
        return pydantic.create_model(name, __base__=base, **fields_constructor)



"""This module provides a high-level interface for interacting with different database clients.

Constants:
    ResourceTable(str): The name of the table in the database.
    ResourceId(uuid.UUID): The unique identifier of the resource.
    ResourceObj(dict[str, typing.Any]): The resource object.
    Partial(bool | None): Perform a partial update.

Classes:
    Base(path: str, table: ResourceTable): Base client class for interacting with a database.
"""

import typing
import uuid

import fastapi
import pydantic

from .. import dependencies

ResourceTable = typing.Annotated[str, pydantic.Field(description="The name of the table in the database")]
ResourceId = typing.Annotated[
    uuid.UUID, pydantic.Field(default_factory=uuid.uuid4, description="The unique identifier of the resource"),
]
ResourceObj = typing.Annotated[
    dict[str, typing.Any], pydantic.Field(default_factory=dict, description="The resource object"),
]
Partial = typing.Annotated[bool | None, pydantic.Field(description="Perform a partial update")]


class Base:
    """Base client class for interacting with a database.

    Provides the basic interface for performing CRUD operations on a database. Derived classes should implement the
    specific methods for each operation.

    Attributes:
        exception (fastapi.HTTPException): The exception module to be used for raising HTTP exceptions.
        status (fastapi.status): The status module to be used for HTTP status codes.

    Args:
        path (str): The path to the database.
        table (ResourceTable): The name of the table in the database.

    Methods:
        insert_one(resource_id: ResourceId, resource_obj: ResourceObj): Inserts a resource object in the database.
            Returns ResourceObj.
        select_one(resource_id: ResourceId): Retrieves a resource from the database based on the given ID. Returns
            ResourceObj.
        update_one(resource_id: ResourceId, resource_obj: ResourceObj, partial: Partial = None): Updates a resource in
            the database based on the given ID. Returns ResourceObj.
        delete_one(resource_id: ResourceId): Deletes a resource from the database based on the given resource ID.
            Returns None.
        select_many(paginate_parameters: dependencies.PaginateParameters): Retrieves multiple resources from the
            database based on the given pagination parameters. Returns list[ResourceObj].
    """

    exception = fastapi.HTTPException
    status = fastapi.status

    def __init__(self, path: str, table: ResourceTable) -> None:
        """Initialize the Base class.

        Args:
            path (str): The path to the database.
            table (ResourceTable): The name of the table in the database.
        """
        self.path = path
        self.table = table

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        raise NotImplementedError

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def update_one(
        self,
        resource_id: ResourceId,
        resource_obj: ResourceObj,
        partial: Partial = None,
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        raise NotImplementedError

    async def select_many(self, paginate_parameters: dependencies.PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        raise NotImplementedError



"""This module provides the client class for interacting with a virtual database.

Classes:
    DefaultClient(_: str, table: str): A class representing a virtual DB client for database operations.
"""

import typing

from . import Base, Partial, ResourceId, ResourceObj, dependencies


class DefaultClient(Base):
    """A class representing a virtual DB client for database operations.

    Attributes:
        db (typing.ClassVar[dict[ResourceId, ResourceObj]]): A dictionary representing the database.

    Args:
        _ (str): A placeholder argument.
        table (str): The name of the table.

    Raises:
        self.exception: If the resource is not found in the database.

    Methods:
        insert_one(resource_id: ResourceId, resource_obj: ResourceObj): Inserts a resource object in the database.
            Returns ResourceObj.
        select_one(resource_id: ResourceId): Retrieves a resource from the database based on the given ID. Returns
            ResourceObj.
        update_one(resource_id: ResourceId, resource_obj: ResourceObj, partial: Partial = None,): Updates a resource in
            the database based on the given ID. Returns ResourceObj.
        delete_one(resource_id: ResourceId): Deletes a resource from the database based on the given resource ID.
            Returns None.
        select_many(paginate_parameters: dependencies.PaginateParameters): Retrieves multiple resources from the
            database based on the given pagination parameters. Returns list[ResourceObj].
    """

    _db: typing.ClassVar[dict[ResourceId, ResourceObj]] = {}

    def __init__(self, _: str, table: str) -> None:
        """Initialize a DefaultClient instance.

        Args:
            _ (str): A placeholder argument.
            table (str): The name of the table.

        Returns:
            None
        """
        super().__init__(_, table)

    def _get(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to retrieve.

        Returns:
            ResourceObj: The retrieved resource.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        try:
            return self._db[resource_id]
        except KeyError as exception:
            raise self.exception(self.status.HTTP_404_NOT_FOUND, "not found") from exception

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        try:
            self._get(resource_id)
        except self.exception:
            self._db[resource_id] = resource_obj
            return {"id": resource_id, **resource_obj}
        else:
            raise self.exception(self.status.HTTP_409_CONFLICT, "conflict")

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        resource_obj = self._get(resource_id)
        return {"id": resource_id, **resource_obj}

    async def update_one(
        self,
        resource_id: ResourceId,
        resource_obj: ResourceObj,
        partial: Partial = None,
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        _resource_obj = self._get(resource_id)
        if partial:
            _resource_obj.update(resource_obj)
        else:
            _resource_obj = resource_obj
        self._db[resource_id] = _resource_obj
        return {"id": resource_id, **_resource_obj}

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        self._get(resource_id)
        self._db.pop(resource_id, None)

    async def select_many(self, paginate_parameters: dependencies.PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        items: list[dict[str, ResourceId | ResourceObj]] = [{"id": k, **v} for k, v in self._db.items()]
        return items[paginate_parameters.skip : paginate_parameters.skip + paginate_parameters.limit]



"""This module provides {...}.

Classes:
    Models: Types class provides a mechanism to manage a custom enumeration of types.

"""

import enum
import typing


class Types:
    """Types class provides a mechanism to manage a custom enumeration of types.

    Attributes:
        types_enum (TypeEnum): An enumeration of various types and values.

    Methods:
        get(name: str) -> typing.Any:
            Retrieve an attribute from the `types_enum` based on the given name.
        set(name: str, value: typing.Any = None, operation: typing.Literal["add", "del"] = "add") -> None:
            Modifies the `types_enum` attribute by adding or deleting an enumeration member.
    """
    class TypeEnum(enum.Enum):
        """Original TypeEnum enumerate class."""

        any = typing.Any
        none = None
        bool = bool
        true = True
        false = False
        str = str
        int = int
        float = float
        list = list
        dict = dict

    types_enum = TypeEnum

    def get(self, name: str) -> typing.Any:
        """Retrieve an attribute from the `types_enum` based on the given name.

        Args:
            name (str): The name of the attribute to retrieve.

        Returns:
            typing.Any: The value of the attribute with the given name.
        """
        return getattr(self.types_enum, name).value

    def set(self, name: str, value: typing.Any = None) -> None:
        """Modifies the `types_enum` attribute by adding an enumeration member.

        Args:
            name (str): The name of the enumeration member to add
            value (typing.Any, optional): The value of the enumeration member to add. Defaults to None.

        Returns:
            None
        """
        old_types_enum: list[tuple[str, typing.Any]] = [
            (i, getattr(self.types_enum, i)) for i in self.types_enum.__members__
        ]
        new_types_enum = [*old_types_enum, (name, value)]
        self.types_enum = enum.Enum("TypeEnum", new_types_enum)

    def pop(self, name: str) -> None:
        """Modifies the `types_enum` attribute by deleting an enumeration member.

        Args:
            name (str): The name of the enumeration member to delete.

        Returns:
            None
        """
        old_types_enum: list[tuple[str, typing.Any]] = [
            (i, getattr(self.types_enum, i)) for i in self.types_enum.__members__
        ]
        new_types_enum = [i for i in old_types_enum if i[0] != name]
        self.types_enum = enum.Enum("TypeEnum", new_types_enum)



import unittest
import unittest.mock

import sthali_db.db

module = sthali_db.db


class TestDBSpecification(unittest.IsolatedAsyncioTestCase):
    async def test_return(self) -> None:
        db_spec = module.DBSpecification(path="test_path", client="tinydb")  # type: ignore

        self.assertEqual(db_spec.path, "test_path")
        self.assertEqual(db_spec.client.name, "tinydb")


class TestDB(unittest.IsolatedAsyncioTestCase):
    class MockDefaultClient:
        insert_one = unittest.mock.AsyncMock(return_value="insert_one")
        select_one = unittest.mock.AsyncMock(return_value="select_one")
        update_one = unittest.mock.AsyncMock(return_value="update_one")
        delete_one = unittest.mock.AsyncMock(return_value="delete_one")
        select_many = unittest.mock.AsyncMock(return_value="select_many")

    @unittest.mock.patch("sthali_db.clients.default.DefaultClient")
    def setUp(self, mocked_client: unittest.mock.MagicMock) -> None:
        mocked_client.return_value = self.MockDefaultClient()
        db_spec = module.DBSpecification("test_path", "default")  # type: ignore
        self.db = module.DB(db_spec, "table")  # type: ignore
        default = module.enum_clients_config.clients_map["default"]
        self.resource_id = default.ResourceId
        self.resource_obj = default.ResourceObj
        self.paginate_parameters = default.dependencies.PaginateParameters

    async def test_insert_one(self) -> None:
        result = await self.db.insert_one(self.resource_id, self.resource_obj)

        self.assertEqual(result, "insert_one")

    async def test_select_one(self) -> None:
        result = await self.db.select_one(self.resource_id)

        self.assertEqual(result, "select_one")

    async def test_update_one(self) -> None:
        result = await self.db.update_one(self.resource_id, self.resource_obj)

        self.assertEqual(result, "update_one")

    async def test_delete_one(self) -> None:
        result = await self.db.delete_one(self.resource_id)

        self.assertEqual(result, "delete_one")

    async def test_select_many(self) -> None:
        result = await self.db.select_many(self.paginate_parameters)

        self.assertEqual(result, "select_many")



"""This module provides the client class for interacting with a TinyDB database."""

import tinydb
import tinydb.table

from . import Base, Partial, ResourceId, ResourceObj, dependencies


class TinydbClient(Base):
    """A class representing a TinyDB client for database operations.

    Args:
        path (str): The path to the TinyDB database file.
        table_name (str): The name of the table in the database.

    Attributes:
        table (tinydb.table.Table): The name of the table in the database.
    """

    table: tinydb.table.Table

    def __init__(self, path: str, table_name: str) -> None:
        """Initialize the TinydbClient class.

        Args:
            path (str): The path to the database.
            table_name (str): The name of the table in the database.

        """
        db = tinydb.TinyDB(path)
        table = db.table(table_name)  # type: ignore

        self.table = table

    def _get(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to retrieve.

        Returns:
            ResourceObj: The retrieved resource.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        try:
            result = self.table.search(tinydb.Query().resource_id == str(resource_id))
            return result[0]["resource_obj"]  # type: ignore
        except (KeyError, IndexError) as exception:
            raise self.exception(self.status.HTTP_404_NOT_FOUND, "not found") from exception

    async def insert_one(self, resource_id: ResourceId, resource_obj: ResourceObj) -> ResourceObj:
        """Inserts a resource object in the database.

        Args:
            resource_id (ResourceId): The ID of the resource to be inserted.
            resource_obj (ResourceObj): The resource object to be inserted.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource already exists in the database.
        """
        try:
            self._get(resource_id)
        except self.exception:
            self.table.insert({"resource_id": str(resource_id), "resource_obj": resource_obj})  # type: ignore
            return {"id": resource_id, **resource_obj}
        else:
            raise self.exception(self.status.HTTP_409_CONFLICT, "conflict")

    async def select_one(self, resource_id: ResourceId) -> ResourceObj:
        """Retrieves a resource from the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be retrieved.

        Returns:
            ResourceObj: The retrieved resource object.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        result = self._get(resource_id)
        return {"id": resource_id, **result}

    async def update_one(
        self,
        resource_id: ResourceId,
        resource_obj: ResourceObj,
        partial: Partial = None,
    ) -> ResourceObj:
        """Updates a resource in the database based on the given ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be updated.
            resource_obj (ResourceObj): The resource object to be updated.
            partial (Partial): Whether to perform a partial update or replace the entire resource object.
                Defaults to None.

        Returns:
            ResourceObj: The resource object containing the ID.

        Raises:
            self.exception: If the resource is not found in the database.
        """
        _resource_obj = self._get(resource_id)
        if partial:
            _resource_obj.update(resource_obj)
        else:
            _resource_obj = resource_obj
        self.table.update({"resource_obj": _resource_obj}, tinydb.Query().resource_id == str(resource_id))  # type: ignore
        return {"id": resource_id, **_resource_obj}

    async def delete_one(self, resource_id: ResourceId) -> None:
        """Deletes a resource from the database based on the given resource ID.

        Args:
            resource_id (ResourceId): The ID of the resource to be deleted.

        Returns:
            None

        Raises:
            self.exception: If the resource is not found in the database.
        """
        self._get(resource_id)
        self.table.remove(tinydb.Query().resource_id == str(resource_id))

    async def select_many(self, paginate_parameters: dependencies.PaginateParameters) -> list[ResourceObj]:
        """Retrieves multiple resources from the database based on the given pagination parameters.

        Args:
            paginate_parameters (PaginateParameters): The pagination parameters.

        Returns:
            list[ResourceObj]: A list of objects representing the retrieved resources.
        """
        return [{"id": result["resource_id"], **result["resource_obj"]} for result in self.table.all()][
            paginate_parameters.skip : paginate_parameters.limit
        ]  # type: ignore



import unittest

import sthali_db.models

module = sthali_db.models

class TestDefault(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = module.FieldSpecification.Default()  # type: ignore

        self.assertEqual(result.factory, None)
        self.assertEqual(result.value, None)

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        result = module.FieldSpecification.Default(func, 0)

        self.assertEqual(result.factory, func)
        self.assertEqual(result.value, 0)


class TestFieldSpecification(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = module.FieldSpecification("test_field_name", module.typing.Any)  # type: ignore

        self.assertEqual(result.name, "test_field_name")
        self.assertEqual(result.type, module.typing.Any)
        self.assertEqual(result.default, None)
        self.assertEqual(result.description, None)
        self.assertEqual(result.optional, None)
        self.assertEqual(result.title, None)

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        optional = True
        result = module.FieldSpecification(
            "test_field_name",
            str,
            {"factory": func, "value": 0},  # type: ignore
            "test_field_description",
            optional,
            "test_field_title",
        )

        self.assertEqual(result.name, "test_field_name")
        self.assertEqual(result.type, str)
        self.assertEqual(result.default.factory, func)  # type: ignore
        self.assertEqual(result.default.value, 0)  # type: ignore
        self.assertEqual(result.description, "test_field_description")
        self.assertEqual(result.optional, True)
        self.assertEqual(result.title, "test_field_title")

    async def test_type_annotated(self) -> None:
        field_spec = module.FieldSpecification("test_field_name", str)  # type: ignore

        result = field_spec.type_annotated

        self.assertEqual(result("str"), "str")
        self.assertEqual(result.__args__[0], str)
        self.assertEqual(result.__metadata__[0].title, "test_field_name")

    async def test_type_annotated_with_optional(self) -> None:
        field_spec = module.FieldSpecification("test_field_name", str, optional=True)  # type: ignore

        result = field_spec.type_annotated

        self.assertEqual(result.__args__[0], str | None)

    async def test_type_annotated_with_default_factory(self) -> None:
        def func() -> None:
            return

        field_spec = module.FieldSpecification(
            "test_field_name", str, {"factory": func, "value": 0}
        )  # type: ignore

        result = field_spec.type_annotated

        self.assertEqual(result.__metadata__[0].default_factory(), None)

    async def test_type_annotated_with_default_value(self) -> None:
        field_spec = module.FieldSpecification("test_field_name", str, {"value": 0})  # type: ignore

        result = field_spec.type_annotated

        self.assertEqual(result.__metadata__[0].default, 0)


class TestBase(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        result = module.Models.Base()

        self.assertEqual(result.model_dump(), {})


class TestBaseWithId(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        _id = module.uuid.uuid4()
        result = module.Models.BaseWithId(id=_id)

        self.assertEqual(result.model_dump(), {"id": _id})


class TestModels(unittest.IsolatedAsyncioTestCase):
    async def test_return_default(self) -> None:
        _id = module.uuid.uuid4()

        result = module.Models("name", [])

        self.assertEqual(result.name, "name")
        self.assertEqual(result.create_model().model_dump(), {})
        self.assertEqual(result.response_model(**{"id": _id}).model_dump(), {"id": _id})
        self.assertEqual(result.update_model().model_dump(), {})

    async def test_return_custom(self) -> None:
        def func() -> None:
            return

        _id = module.uuid.uuid4()

        result = module.Models(
            name="name",
            fields=[
                module.FieldSpecification("test_field_name_1", str),  # type: ignore
                module.FieldSpecification("test_field_name_2", str, optional=True),  # type: ignore
                module.FieldSpecification("test_field_name_3", str, default={"factory": func}),  # type: ignore
                module.FieldSpecification("test_field_name_4", str, default={"value": "test_field_value_4"}),  # type: ignore
                module.FieldSpecification("test_field_name_5", str, default={"factory": func, "value": 1}),  # type: ignore
            ],
        )

        self.assertEqual(result.name, "name")
        self.assertEqual(
            result.create_model(**{"test_field_name_1": "test_field_value_1", "test_field_name_2": None}).model_dump(),
            {
                "test_field_name_1": "test_field_value_1",
                "test_field_name_2": None,
                "test_field_name_3": None,
                "test_field_name_4": "test_field_value_4",
                "test_field_name_5": None,
            },
        )
        self.assertEqual(
            result.response_model(
                **{"id": _id, "test_field_name_1": "test_field_value_1", "test_field_name_2": None}  # type: ignore
            ).model_dump(),
            {
                "id": _id,
                "test_field_name_1": "test_field_value_1",
                "test_field_name_2": None,
                "test_field_name_3": None,
                "test_field_name_4": "test_field_value_4",
                "test_field_name_5": None,
            },
        )
        self.assertEqual(
            result.update_model(**{"test_field_name_1": "test_field_value_1", "test_field_name_2": None}).model_dump(),
            {
                "test_field_name_1": "test_field_value_1",
                "test_field_name_2": None,
                "test_field_name_3": None,
                "test_field_name_4": "test_field_value_4",
                "test_field_name_5": None,
            },
        )
